@@ -36,19 +36,7 @@
 #include "sym.c"
 #include "regs.h"
 
-/* If uprobes isn't in the kernel, pull it in from the runtime. */
-#if defined(CONFIG_UTRACE)      /* uprobes doesn't work without utrace */
-#if defined(CONFIG_UPROBES) || defined(CONFIG_UPROBES_MODULE)
-#include <linux/uprobes.h>
-#else
-#include "uprobes/uprobes.h"
-#endif
-#ifndef UPROBES_API_VERSION
-#define UPROBES_API_VERSION 1
-#endif
-#else
-struct uretprobe_instance;
-#endif
+#include "uprobes-inc.h"
 
 #if defined(STAPCONF_KERNEL_STACKTRACE) || defined(STAPCONF_KERNEL_STACKTRACE_NO_BP)
 #include <linux/stacktrace.h>
