@@ -13,33 +13,19 @@
 #ifndef _STP_UTRACE_C
 #define _STP_UTRACE_C
 
-/*
- * Which internal utrace implementation shall we use?
- * (1) STAPCONF_UTRACE_VIA_TRACEPOINTS - Pure tracepoints (preferred).
- * (2) STAPCONF_UTRACE_VIA_FTRACE - Tracepoints, plus a bit of ftrace.
- */
-
-#if (!defined(STAPCONF_UTRACE_VIA_TRACEPOINTS) \
-     && !defined(STAPCONF_UTRACE_VIA_FTRACE))
-#error "STAPCONF_UTRACE_VIA_TRACEPOINTS or STAPCONF_UTRACE_VIA_FTRACE must be defined."
+#if (!defined(STAPCONF_UTRACE_VIA_TRACEPOINTS))
+#error "STAPCONF_UTRACE_VIA_TRACEPOINTS must be defined."
 #endif
-#if (defined(STAPCONF_UTRACE_VIA_TRACEPOINTS) \
-     && defined(STAPCONF_UTRACE_VIA_FTRACE))
-/* If we've got both, prefer STAPCONF_UTRACE_VIA_TRACEPOINTS. */
-#undef STAPCONF_UTRACE_VIA_FTRACE
-#endif /* STAPCONF_UTRACE_VIA_TRACEPOINTS and STAPCONF_UTRACE_VIA_FTRACE */
 
 #include "stp_utrace.h"
 #include <linux/list.h>
 #include <linux/sched.h>
 #include <linux/freezer.h>
 #include <linux/slab.h>
-#ifdef STAPCONF_UTRACE_VIA_FTRACE
-#include <linux/ftrace.h>
-#endif	/* STAPCONF_UTRACE_VIA_FTRACE */
 #include <linux/spinlock.h>
 #include <trace/events/sched.h>
 #include <trace/events/syscalls.h>
+#include <linux/task_work.h>
 
 /*
  * Per-thread structure private to utrace implementation.
@@ -79,11 +65,14 @@ struct utrace {
 	unsigned int death:1;	/* in utrace_report_death() now */
 	unsigned int reap:1;	/* release_task() has run */
 	unsigned int pending_attach:1; /* need splice_attaching() */
+	unsigned int task_work_added:1; /* called task_work_add() */
 
 	unsigned long utrace_flags;
 
 	struct hlist_node hlist;       /* task_utrace_table linkage */
 	struct task_struct *task;
+
+	struct task_work work;
 };
 
 #define TASK_UTRACE_HASH_BITS 5
@@ -107,31 +96,46 @@ static void utrace_report_syscall_entry(void *cb_data __attribute__ ((unused)),
 static void utrace_report_syscall_exit(void *cb_data __attribute__ ((unused)),
 				       struct pt_regs *regs, long ret);
 
-#ifdef STAPCONF_UTRACE_VIA_TRACEPOINTS
 static void utrace_report_exec(void *cb_data __attribute__ ((unused)),
 			       struct task_struct *task,
 			       pid_t old_pid __attribute__((unused)),
 			       struct linux_binprm *bprm __attribute__ ((unused)));
-#else  /* STAPCONF_UTRACE_VIA_FTRACE */
-static void utrace_report_exec(unsigned long ip __attribute__ ((unused)),
-			       unsigned long parent_ip __attribute__ ((unused)));
-
-static struct ftrace_ops utrace_report_exec_ops __read_mostly =
-{
-    .func = utrace_report_exec,
-};
-#endif  /* STAPCONF_UTRACE_VIA_FTRACE */
 
 #define __UTRACE_UNREGISTERED	0
 #define __UTRACE_REGISTERED	1
 static atomic_t utrace_state = ATOMIC_INIT(__UTRACE_UNREGISTERED);
 
+#if !defined(STAPCONF_TASK_WORK_ADD_EXPORTED)
+typedef int (*task_work_add_fn)(struct task_struct *task,
+				  struct task_work *twork, bool notify);
+#define task_work_add (* (task_work_add_fn)kallsyms_task_work_add)
+typedef struct task_work *(*task_work_cancel_fn)(struct task_struct *,
+						 task_work_func_t);
+#define task_work_cancel (* (task_work_cancel_fn)kallsyms_task_work_cancel)
+#endif
+
 int utrace_init(void)
 {
 	int i;
 	int rc = -1;
-#ifdef STAPCONF_UTRACE_VIA_FTRACE
-	char *report_exec_name;
+
+#if !defined(STAPCONF_TASK_WORK_ADD_EXPORTED)
+	/* The task_work_add()/task_work_cancel() functions aren't
+	 * exported. Look up those function addresses. */
+        kallsyms_task_work_add = (void *)kallsyms_lookup_name ("task_work_add");
+        if (kallsyms_task_work_add == NULL) {
+		printk(KERN_ERR "%s can't resolve task_work_add!",
+		       THIS_MODULE->name);
+		rc = -ENOENT;
+		goto error;
+        }
+        kallsyms_task_work_cancel = (void *)kallsyms_lookup_name ("task_work_cancel");
+        if (kallsyms_task_work_cancel == NULL) {
+		printk(KERN_ERR "%s can't resolve task_work_cancel!",
+		       THIS_MODULE->name);
+		rc = -ENOENT;
+		goto error;
+        }
 #endif
 
 	/* initialize the list heads */
@@ -167,25 +171,13 @@ int utrace_init(void)
 		goto error4;
 	}
 
-#ifdef STAPCONF_UTRACE_VIA_TRACEPOINTS
 	rc = register_trace_sched_process_exec(utrace_report_exec, NULL);
 	if (unlikely(rc != 0)) {
 		_stp_error("register_sched_process_exec failed: %d", rc);
 		goto error5;
 	}
-#else  /* STAPCONF_UTRACE_VIA_FTRACE */
-	report_exec_name = "*" __stringify(proc_exec_connector);
-	ftrace_set_filter(&utrace_report_exec_ops, report_exec_name,
-			  strlen(report_exec_name), 1);
-	rc = register_ftrace_function(&utrace_report_exec_ops);
-	if (unlikely(rc != 0)) {
-		_stp_error("register_ftrace_function failed: %d", rc);
-		goto error5;
-	}
-#endif  /* STAPCONF_UTRACE_VIA_FTRACE */
 
 	atomic_set(&utrace_state, __UTRACE_REGISTERED);
-
 	return 0;
 
 error5:
@@ -216,6 +208,8 @@ int utrace_exit(void)
 	return 0;
 }
 
+void utrace_resume(struct task_work *work);
+
 /*
  * Clean up everything associated with @task.utrace.
  *
@@ -243,6 +237,16 @@ static void utrace_cleanup(struct utrace *utrace)
 	    list_del(&engine->entry);
 	    kmem_cache_free(utrace_engine_cachep, engine);
 	}
+
+	if (utrace->task_work_added) {
+		if (task_work_cancel(utrace->task, &utrace_resume) == NULL)
+			printk(KERN_ERR "%s:%d - task_work_cancel() failed? task %p, %d, %s\n",
+			       __FUNCTION__, __LINE__, utrace->task,
+			       utrace->task->tgid,
+			       (utrace->task->comm ? utrace->task->comm
+				: "UNKNOWN"));
+		utrace->task_work_added = 0;
+	}
 	spin_unlock(&utrace->lock);
 
 	/* Free the struct utrace itself. */
@@ -265,11 +269,7 @@ void utrace_shutdown(void)
 #ifdef STP_TF_DEBUG
 	printk(KERN_ERR "%s:%d entry\n", __FUNCTION__, __LINE__);
 #endif
-#ifdef STAPCONF_UTRACE_VIA_TRACEPOINTS
 	unregister_trace_sched_process_exec(utrace_report_exec, NULL);
-#else  /* STAPCONF_UTRACE_VIA_FTRACE */
-	unregister_ftrace_function(&utrace_report_exec_ops);
-#endif  /* STAPCONF_UTRACE_VIA_FTRACE */
 	unregister_trace_sched_process_fork(utrace_report_clone, NULL);
 	unregister_trace_sched_process_exit(utrace_report_death, NULL);
 	unregister_trace_sys_enter(utrace_report_syscall_entry, NULL);
@@ -345,6 +345,7 @@ static bool utrace_task_alloc(struct task_struct *task)
 		kmem_cache_free(utrace_cachep, utrace);
 	}
 
+	init_task_work(&utrace->work, &utrace_resume, NULL);
 	return true;
 }
 
@@ -376,6 +377,17 @@ static void utrace_free(struct utrace *utrace)
 		       list_empty(&utrace->attached),
 		       list_empty(&utrace->attaching));
 #endif
+
+	if (utrace->task_work_added) {
+		if (task_work_cancel(utrace->task, &utrace_resume) == NULL)
+			printk(KERN_ERR "%s:%d - task_work_cancel() failed? task %p, %d, %s\n",
+			       __FUNCTION__, __LINE__, utrace->task,
+			       utrace->task->tgid,
+			       (utrace->task->comm ? utrace->task->comm
+				: "UNKNOWN"));
+		utrace->task_work_added = 0;
+	}
+
 	kmem_cache_free(utrace_cachep, utrace);
 }
 
@@ -883,12 +895,22 @@ static bool utrace_do_stop(struct task_struct *target, struct utrace *utrace)
 		if (likely(task_is_stopped(target)))
 			__set_task_state(target, TASK_TRACED);
 		spin_unlock_irq(&target->sighand->siglock);
-#if 0
-	/* FIXME: needed?  If so, what to do here? */
 	} else if (utrace->resume > UTRACE_REPORT) {
 		utrace->resume = UTRACE_REPORT;
-		set_notify_resume(target);
-#endif
+		if (! utrace->task_work_added) {
+			int rc = task_work_add(target, &utrace->work, true);
+			if (rc == 0) {
+				utrace->task_work_added = 1;
+			}
+			/* task_work_add() returns -ESRCH if the task
+			 * has already passed exit_task_work(). Just
+			 * ignore this error. */
+			else if (rc != -ESRCH) {
+				printk(KERN_ERR
+				       "%s:%d - task_work_add() returned %d\n",
+				       __FUNCTION__, __LINE__, rc);
+			}
+		}
 	}
 
 	return task_is_traced(target);
@@ -1018,8 +1040,20 @@ relock:
 		 * Ensure a reporting pass when we're resumed.
 		 */
 		utrace->resume = action;
-		/* FIXME: needed? */
-		set_thread_flag(TIF_NOTIFY_RESUME);
+		if (! utrace->task_work_added) {
+			int rc = task_work_add(task, &utrace->work, true);
+			if (rc == 0) {
+				utrace->task_work_added = 1;
+			}
+			/* task_work_add() returns -ESRCH if the task
+			 * has already passed exit_task_work(). Just
+			 * ignore this error. */
+			else if (rc != -ESRCH) {
+				printk(KERN_ERR
+				       "%s:%d - task_work_add() returned %d\n",
+				       __FUNCTION__, __LINE__, rc);
+			}
+		}
 	}
 
 	/*
@@ -1360,13 +1394,25 @@ int utrace_control(struct task_struct *target,
 		 * In that case, utrace_get_signal() will be reporting soon.
 		 */
 		clear_engine_wants_stop(engine);
-#if 0
-		/* FIXME: needed?  If so, what to do here? */
 		if (action < utrace->resume) {
 			utrace->resume = action;
-			set_notify_resume(target);
+			if (! utrace->task_work_added) {
+				ret = task_work_add(target, &utrace->work,
+						    true);
+				if (ret == 0) {
+					utrace->task_work_added = 1;
+				}
+				/* task_work_add() returns -ESRCH if
+				 * the task has already passed
+				 * exit_task_work(). Just ignore this
+				 * error. */
+				else if (ret != -ESRCH) {
+					printk(KERN_ERR
+					       "%s:%d - task_work_add() returned %d\n",
+					       __FUNCTION__, __LINE__, ret);
+				}
+			}
 		}
-#endif
 		break;
 
 	default:
@@ -1513,8 +1559,20 @@ static void finish_report(struct task_struct *task, struct utrace *utrace,
 	if (resume < utrace->resume) {
 		spin_lock(&utrace->lock);
 		utrace->resume = resume;
-		/* FIXME: Hmm, unsure about calling set_tsk_thread_flag()... */
-		set_tsk_thread_flag(task, TIF_NOTIFY_RESUME);
+		if (! utrace->task_work_added) {
+			int rc = task_work_add(task, &utrace->work, true);
+			if (rc == 0) {
+				utrace->task_work_added = 1;
+			}
+			/* task_work_add() returns -ESRCH if the task
+			 * has already passed exit_task_work(). Just
+			 * ignore this error. */
+			else if (rc != -ESRCH) {
+				printk(KERN_ERR
+				       "%s:%d - task_work_add() returned %d\n",
+				       __FUNCTION__, __LINE__, rc);
+			}
+		}
 		spin_unlock(&utrace->lock);
 	}
 
@@ -1732,19 +1790,11 @@ static const struct utrace_engine_ops *start_callback(
 /*
  * Called iff UTRACE_EVENT(EXEC) flag is set.
  */
-#ifdef STAPCONF_UTRACE_VIA_TRACEPOINTS
 static void utrace_report_exec(void *cb_data __attribute__ ((unused)),
 			       struct task_struct *task,
 			       pid_t old_pid __attribute__((unused)),
 			       struct linux_binprm *bprm __attribute__ ((unused)))
-#else  /* STAPCONF_UTRACE_VIA_FTRACE */
-static void utrace_report_exec(unsigned long ip __attribute__ ((unused)),
-			       unsigned long parent_ip __attribute__ ((unused)))
-#endif  /* STAPCONF_UTRACE_VIA_FTRACE */
 {
-#ifdef STAPCONF_UTRACE_VIA_FTRACE
-	struct task_struct *task = current;
-#endif
 	struct utrace *utrace = task_utrace_struct(task);
 
 	if (utrace && utrace->utrace_flags & UTRACE_EVENT(EXEC)) {
@@ -1883,7 +1933,6 @@ static void utrace_report_clone(void *cb_data __attribute__ ((unused)),
 		unsigned long clone_flags = 0;
 		INIT_REPORT(report);
 
-
 		/* FIXME: Figure out what the clone_flags were. For
 		 * task_finder's purposes, all we need is CLONE_THREAD. */
 		if (task->mm == child->mm)
@@ -2047,12 +2096,22 @@ static void finish_resume_report(struct task_struct *task,
  * We are close to user mode, and this is the place to report or stop.
  * When we return, we're going to user mode or into the signals code.
  */
-void utrace_resume(struct task_struct *task, struct pt_regs *regs)
+void utrace_resume(struct task_work *work)
 {
-	struct utrace *utrace = task_utrace_struct(task);
+	/*
+	 * We could also do 'task_utrace_struct()' here to find the
+	 * task's 'struct utrace', but 'container_of()' should be
+	 * instantaneous (where 'task_utrace_struct()' has to do a
+	 * hash lookup).
+	 */
+	struct utrace *utrace = container_of(work, struct utrace, work);
+	struct task_struct *task = current;
 	INIT_REPORT(report);
 	struct utrace_engine *engine;
 
+	might_sleep();
+	utrace->task_work_added = 0;
+
 	/*
 	 * Some machines get here with interrupts disabled.  The same arch
 	 * code path leads to calling into get_signal_to_deliver(), which
