@@ -38,6 +38,9 @@ extern "C" {
 #include <unistd.h>
 #include <regex.h>
 #include <stdarg.h>
+#ifndef SINGLE_THREADED
+#include <pthread.h>
+#endif
 }
 
 using namespace std;
@@ -56,8 +59,8 @@ get_home_directory(void)
   if (pwd)
     return pwd->pw_dir;
 
-  throw runtime_error(_("Unable to determine home directory"));
-  return NULL;
+  cerr << _("Unable to determine home directory") << endl;
+  return "/";
 }
 
 
@@ -280,12 +283,24 @@ getmemusage ()
   ostringstream oss;
   ifstream statm("/proc/self/statm");
   statm >> pages;
-  long kb1 = pages * sz / 1024;
+  long kb1 = pages * sz / 1024; // total program size; vmsize
   statm >> pages;
-  long kb2 = pages * sz / 1024;
+  long kb2 = pages * sz / 1024; // resident set size; vmrss
   statm >> pages;
-  long kb3 = pages * sz / 1024;
-  oss << _F("using %ldvirt/%ldres/%ldshr kb, ", kb1, kb2, kb3);
+  long kb3 = pages * sz / 1024; // shared pages
+  statm >> pages;
+  long kb4 = pages * sz / 1024; // text
+  statm >> pages;
+  (void) kb4;
+  long kb5 = pages * sz / 1024; // library
+  statm >> pages;
+  (void) kb5;
+  long kb6 = pages * sz / 1024; // data+stack
+  statm >> pages;
+  long kb7 = pages * sz / 1024; // dirty
+  (void) kb7;
+
+  oss << _F("using %ldvirt/%ldres/%ldshr/%lddata kb, ", kb1, kb2, kb3, kb6);
   return oss.str();
 }
 
@@ -309,6 +324,47 @@ tokenize(const string& str, vector<string>& tokens,
     }
 }
 
+// Akin to tokenize(...,...), but allow tokens before the first delimeter, after the
+// last delimiter and allow internal empty tokens
+void
+tokenize_full(const string& str, vector<string>& tokens,
+	      const string& delimiters = " ")
+{
+  // Check for an empty string or a string of length 1. Neither can have the requested
+  // components.
+  if (str.size() <= 1)
+    return;
+
+  // Find the first delimeter.
+  string::size_type lastPos = 0;
+  string::size_type pos = str.find_first_of(delimiters, lastPos);
+  if (pos == string::npos)
+    return; // no delimeters
+
+  /* No leading empty component allowed. */
+  if (pos == lastPos)
+    ++lastPos;
+
+  assert (lastPos < str.size());
+  do
+    {
+      pos = str.find_first_of(delimiters, lastPos);
+      if (pos == string::npos)
+	break; // Final trailing component
+      // Found a token, add it to the vector.
+      tokens.push_back(str.substr (lastPos, pos - lastPos));
+      // Skip the delimiter.
+      lastPos = pos + 1;
+    }
+  while (lastPos < str.size());
+
+  // A final non-delimited token, if it is not empty.
+  if (lastPos < str.size())
+    {
+      assert (pos == string::npos);
+      tokens.push_back(str.substr (lastPos));
+    }
+}
 
 // Akin to tokenize(...,"::"), but it also has to deal with C++ template
 // madness.  We do this naively by balancing '<' and '>' characters.  This
@@ -345,7 +401,9 @@ tokenize_cxx(const string& str, vector<string>& tokens)
 // same policy as execvp().  A program name not containing a slash
 // will be searched along the $PATH.
 
-string find_executable(const string& name, const string& env_path)
+string find_executable(const string& name, const string& sysroot,
+		       map<string, string>& sysenv,
+		       const string& env_path)
 {
   string retpath;
 
@@ -356,11 +414,15 @@ string find_executable(const string& name, const string& env_path)
 
   if (name.find('/') != string::npos) // slash in the path already?
     {
-      retpath = name;
+      retpath = sysroot + name;
     }
   else // Nope, search $PATH.
     {
-      char *path = getenv(env_path.c_str());
+      char *path;
+      if (sysenv.count(env_path) != 0)
+        path = (char *)sysenv[env_path].c_str();
+      else
+        path = getenv(env_path.c_str());
       if (path)
         {
           // Split PATH up.
@@ -370,7 +432,7 @@ string find_executable(const string& name, const string& env_path)
           // Search the path looking for the first executable of the right name.
           for (vector<string>::iterator i = dirs.begin(); i != dirs.end(); i++)
             {
-              string fname = *i + "/" + name;
+              string fname = sysroot + *i + "/" + name;
               const char *f = fname.c_str();
 
               // Look for a normal executable file.
@@ -389,13 +451,22 @@ string find_executable(const string& name, const string& env_path)
   // Could not find the program on the $PATH.  We'll just fall back to
   // the unqualified name, which our caller will probably fail with.
   if (retpath == "")
-    retpath = name;
+    retpath = sysroot + name;
 
   // Canonicalize the path name.
   char *cf = canonicalize_file_name (retpath.c_str());
   if (cf)
     {
-      retpath = string(cf);
+      string scf = string(cf);
+      if (sysroot.empty())
+        retpath = scf;
+      else {
+        int pos = scf.find(sysroot);
+        if (pos == 0)
+          retpath = scf;
+        else
+          throw runtime_error(_F("find_executable(): file %s not in sysroot %s", cf, sysroot.c_str()));
+      }
       free (cf);
     }
 
@@ -454,32 +525,80 @@ cmdstr_join(const vector<string>& cmds)
 class spawned_pids_t {
   private:
     set<pid_t> pids;
+#ifndef SINGLE_THREADED
+    pthread_mutex_t mux_pids;
+#endif
 
   public:
     bool contains (pid_t p)
       {
         stap_sigmasker masked;
-        return pids.count(p) == 0;
+
+#ifndef SINGLE_THREADED
+        pthread_mutex_lock(&mux_pids);
+#endif
+        bool ret = (pids.count(p)==0) ? true : false;
+#ifndef SINGLE_THREADED
+        pthread_mutex_unlock(&mux_pids);
+#endif
+
+        return ret;
       }
     bool insert (pid_t p)
       {
         stap_sigmasker masked;
-        return (p > 0) ? pids.insert(p).second : false;
+
+#ifndef SINGLE_THREADED
+        pthread_mutex_lock(&mux_pids);
+#endif
+        bool ret = (p > 0) ? pids.insert(p).second : false;
+#ifndef SINGLE_THREADED
+        pthread_mutex_unlock(&mux_pids);
+#endif
+
+        return ret;
       }
     void erase (pid_t p)
       {
         stap_sigmasker masked;
+
+#ifndef SINGLE_THREADED
+        pthread_mutex_lock(&mux_pids);
+#endif
         pids.erase(p);
+#ifndef SINGLE_THREADED
+        pthread_mutex_unlock(&mux_pids);
+#endif
       }
     int killall (int sig)
       {
         int ret = 0;
         stap_sigmasker masked;
+
+#ifndef SINGLE_THREADED
+        pthread_mutex_lock(&mux_pids);
+#endif
         for (set<pid_t>::const_iterator it = pids.begin();
              it != pids.end(); ++it)
           ret = kill(*it, sig) ?: ret;
+#ifndef SINGLE_THREADED
+        pthread_mutex_unlock(&mux_pids);
+#endif
         return ret;
       }
+    spawned_pids_t()
+      {
+#ifndef SINGLE_THREADED
+        pthread_mutex_init(&mux_pids, NULL);
+#endif
+      }
+    ~spawned_pids_t()
+      {
+#ifndef SINGLE_THREADED
+        pthread_mutex_destroy (&mux_pids);
+#endif
+      }
+
 };
 static spawned_pids_t spawned_pids;
 
@@ -678,7 +797,8 @@ localization_variables()
 // Runs a command with a saved PID, so we can kill it from the signal handler,
 // and wait for it to finish.
 int
-stap_system(int verbose, const vector<string>& args,
+stap_system(int verbose, const string& description,
+            const vector<string>& args,
             bool null_out, bool null_err)
 {
   int ret = 0;
@@ -697,7 +817,7 @@ stap_system(int verbose, const vector<string>& args,
         ret = stap_waitpid(verbose, pid);
         if(ret)
           // XXX PR13274 needs-session to use print_warning()
-          clog << _F("WARNING: %s exited with status: %d", args.front().c_str(), ret) << endl;
+          clog << _F("WARNING: %s exited with status: %d", description.c_str(), ret) << endl;
       }
     }
 
@@ -908,7 +1028,7 @@ std::string autosprintf(const char* format, ...)
   va_start (args, format);
   int rc = vasprintf (&str, format, args);
   if (rc < 0)
-    throw runtime_error (_F("autosprintf/vasprintf error %s", lex_cast(rc).c_str()));
+    return _F("autosprintf/vasprintf error %s", lex_cast(rc).c_str());
   string s = str;
   va_end (args);
   free (str);
